# -*- coding: UTF-8 -*-
import os
import copy
from collections import Iterable
from decimal import Decimal


def get_chunk(l, n):
    """
    get a chunk in iterable object
    :param l: iterable object
    :param n: chunk size
    :return: a chunk in list type
    """
    if not isinstance(l, Iterable):
        raise TypeError('input value is not iterable')
    if hasattr(l, '__getitem__'):
        for i in range(0, len(l), n):
            yield l[i:i + n]
    else:
        chunk = []
        for i in l:
            chunk.append(i)
            if len(chunk) == n:
                yield chunk
                chunk = []

        if chunk:
            yield chunk


def list2dict(l, key, pop_key=False):
    """
    convert list of dict to dict
    :param l: list of dict
    :param key: key name, which value of dict in list as returned dict key
    :param pop_key: whether pop the key in list of
    :return: assembled dict
    """
    if type(l) not in {list, tuple}:
        raise TypeError('input must in list or tuple type')
    d = {}
    for item in l:
        if not isinstance(item, dict):
            raise TypeError('item {0} is not dict'.format(item))
        if key not in item:
            raise KeyError('key is not in item')
        new_item = copy.deepcopy(item)
        if pop_key:
            new_item.pop(key)
        d[item[key]] = new_item
    return d


def get_filenames_in_dir(dirname,
                         skip_dir=True,
                         skip_hidden_file=True,
                         rm_extname=False,
                         rm_dirname=False):
    """
    get file names in directory, file name in dictionary order
    :param dirname: directory path to get file names
    :param skip_dir: whether skip directory in results
    :param skip_hidden_file: whether skip hidden file in results
    :param rm_extname: whether remove suffix extname in results
    :param rm_dirname: whether remove prefix dirname in results
    :return: scanned file name list, in dictionary order
    """
    filenames = []
    if not os.path.exists(dirname):
        raise FileNotFoundError('directory is not existed.')

    for filename in os.listdir(dirname):
        is_add = True
        if skip_hidden_file and filename.startswith('.') and filename not in {'.', '..'}:
            is_add = False
        if skip_dir and os.path.isdir(os.path.join(dirname, filename)):
            is_add = False
        if is_add:
            filenames.append(filename)

    if not rm_dirname:
        filenames = [os.path.join(dirname, name) for name in filenames]

    if rm_extname:
        new_filenames = []
        for filename in filenames:
            basename = os.path.basename(filename)
            if '.' in basename:
                dot_index = basename.rindex('.')
                if dot_index:
                    filename = filename[:filename.rindex('.')]

            new_filenames.append(filename)
        filenames = new_filenames

    filenames = sorted(filenames)
    return filenames


def index(l, val, default=-1):
    """
    find the index the val in list
    :param l: index list
    :param val: value to find index
    :param default: default value to return that value not in list
    :return: value index in list
    """
    if type(l) not in {list, tuple}:
        raise TypeError('value is not in index')
    if val not in l:
        return default
    else:
        return l.index(val)


def json_serialize(obj):
    """
    add serialize method used in json.dumps or json.dump
    :param obj: input obj
    :return: string representation for json.dump or json.dumps
    """
    if isinstance(obj, Decimal):
        numerator, denumerator = obj.as_integer_ratio()
        if denumerator == 1:
            return str(numerator)
        else:
            return str(obj)
    elif isinstance(obj, bytes):
        return str(obj)
    else:
        try:
            return str(obj)
        except:
            raise TypeError(repr(obj) + ' is not JSON serializable')



# -*- coding: UTF-8 -*-
import shutil
import tempfile
from types import GeneratorType
from decimal import Decimal
import pytest
from pysenal.utils.utils import *


def test_get_chunk():
    r1 = range(100)
    chunk_r1 = get_chunk(r1, 12)
    assert type(chunk_r1) == GeneratorType
    assert next(chunk_r1) == range(12)

    l1 = [1, 3, 5, 7, 9]
    chunk_l1 = get_chunk(l1, 5)
    assert next(chunk_l1) == l1
    with pytest.raises(StopIteration):
        next(chunk_l1)

    chunk_l1 = get_chunk(l1, 10)
    assert next(chunk_l1) == l1

    g1 = (i for i in [1, 10, 100, 3, 4, 'a'])
    chunk_g1 = get_chunk(g1, 2)
    assert type(chunk_g1) == GeneratorType
    assert next(chunk_g1) == [1, 10]
    assert next(chunk_g1) == [100, 3]

    g2 = (i for i in [20, 100, 40, 123])
    chunk_g2 = get_chunk(g2, 3)
    assert next(chunk_g2) == [20, 100, 40]
    assert next(chunk_g2) == [123]

    g3 = (i for i in [10, 100, 200])
    chunk_g3 = get_chunk(g3, 5)
    assert next(chunk_g3, [10, 100, 200])
    with pytest.raises(StopIteration):
        next(chunk_g3)


def test_list2dict():
    l1 = [{'pid': '1', 'title': 'haha'}, {'pid': '2', 'title': 'lalala'}]
    expected_l1 = {'1': {'pid': '1', 'title': 'haha'}, '2': {'pid': '2', 'title': 'lalala'}}
    ret_l1 = list2dict(l1, 'pid')
    l1[0]['pid'] = '11'
    l1[1]['pid'] = '22'

    assert ret_l1 == expected_l1
    with pytest.raises(KeyError):
        list2dict(l1, 'patent_id')
    l1.append(1)
    with pytest.raises(TypeError):
        list2dict(l1, 'pid')

    l2 = [{'name': 'a', 'count': 11}, {'name': 'b', 'count': 2}]
    expected_l2 = {'a': {'count': 11}, 'b': {'count': 2}}
    assert list2dict(l2, 'name', pop_key=True) == expected_l2
    with pytest.raises(TypeError):
        list2dict(1, 'pid')


@pytest.fixture()
def get_filename_dirname():
    base_dirname = tempfile.gettempdir() + '/'
    dirname = base_dirname + 'pysenal_test_utils_get_filename/'
    return dirname


@pytest.fixture()
def get_filename_dir_setup(get_filename_dirname):
    dirname = get_filename_dirname
    if os.path.exists(dirname):
        shutil.rmtree(dirname)
    os.mkdir(dirname)
    os.mkdir(dirname + 'dir1')
    os.mkdir(dirname + 'dir2')
    open(dirname + 'file1', 'a').close()
    open(dirname + 'file2.bak', 'a').close()
    open(dirname + 'file2.bak', 'a').close()
    open(dirname + '.file3', 'a').close()
    open(dirname + '.file4.txt', 'a').close()


def test_get_filename_in_dir(get_filename_dir_setup, get_filename_dirname):
    dirname = get_filename_dirname
    expected_ret1 = [dirname + 'file1', dirname + 'file2.bak']
    assert get_filenames_in_dir(dirname) == expected_ret1

    expected_ret2 = [dirname + 'file1', dirname + 'file2']
    assert get_filenames_in_dir(dirname, rm_extname=True) == expected_ret2

    assert get_filenames_in_dir(dirname, rm_extname=True, rm_dirname=True) == ['file1', 'file2']

    ret3 = get_filenames_in_dir(dirname, rm_extname=True, skip_hidden_file=False)
    expected_ret3 = [dirname + '.file3', dirname + '.file4', dirname + 'file1', dirname + 'file2']
    assert ret3 == expected_ret3


def test_index():
    values = [1, 2, 3, 4, 5]
    assert index(values, 2) == 1
    assert index(values, 6) == -1
    assert index(values, 6, default=100) == 100
    with pytest.raises(TypeError):
        index(10, 10)


def test_json_serialize():
    assert json_serialize(Decimal('3.1415926')) == '3.1415926'
    assert json_serialize(Decimal('11.0')) == '11'
    assert json_serialize(b'123') == "b'123'"
    assert json_serialize(11) == '11'



# -*- coding: UTF-8 -*-
"""
io related utils functions
"""
import json
import os
from collections import Iterable
import configparser
from ..utils.logger import get_logger

_ENCODING_UTF8 = 'utf-8'

_LINE_BREAKS = '\n\v\x0b\f\x0c\x1c\x1d\x1e\x85\u2028\u2029'
_LINE_BREAK_TUPLE = tuple(_LINE_BREAKS)


def read_lines(filename, encoding=_ENCODING_UTF8, keep_end=False,
               strip=False, skip_empty=False, default=None):
    """
    read lines in text file
    :param filename: file path
    :param encoding: encoding of the file, default is utf-8
    :param keep_end: whether keep line break in result lines
    :param strip: whether strip every line, default is False
    :param skip_empty: whether skip empty line, when strip is False, judge after strip
    :param default: returned value when filename is not existed.
                    If it's None, exception will be raised as usual.
    :return: lines
    """
    if not os.path.exists(filename) and default is not None:
        return default
    with open(filename, encoding=encoding) as f:
        if strip:
            if skip_empty:
                return [l.strip() for l in f.read().splitlines() if l.strip()]
            else:
                return [l.strip() for l in f.read().splitlines()]
        else:
            if skip_empty:
                lines = []
                for line in f.read().splitlines(True):
                    if not keep_end:
                        line = line.rstrip(_LINE_BREAKS)
                    if line:
                        lines.append(line)
                return lines
            else:
                if not keep_end:
                    return [l.rstrip(_LINE_BREAKS) for l in f.read().splitlines(True)]
                else:
                    return f.read().splitlines(True)


def read_lines_lazy(filename, encoding=_ENCODING_UTF8, keep_end=False,
                    strip=False, skip_empty=False, default=None):
    """
    use generator to load files, one line every time
    :param filename: source file path
    :param encoding: file encoding
    :param keep_end: whether keep line break in result lines
    :param strip: whether strip every line, default is False
    :param skip_empty: whether skip empty line, when strip is False, judge after strip
    :param default: returned value when filename is not existed.
                    If it's None, exception will be raised as usual.
    :return: lines in file one by one
    """
    if not os.path.exists(filename) and default is not None:
        return default
    file = open(filename, encoding=encoding)
    for line in file:
        if not keep_end:
            line = line.rstrip(_LINE_BREAKS)
        if strip:
            line = line.strip()
        if skip_empty and not line:
            continue
        yield line
    file.close()


def read_file(filename, encoding=_ENCODING_UTF8, default=None):
    """
    wrap open function to read text in file
    :param filename: file path
    :param encoding: encoding of file, default is utf-8
    :param default: returned value when filename is not existed.
                    If it's None, exception will be raised as usual.
    :return: text in file
    """
    if not os.path.exists(filename) and default is not None:
        return default
    with open(filename, encoding=encoding) as f:
        return f.read()


def write_file(filename, data, encoding=_ENCODING_UTF8):
    """
    write text into file
    :param filename: file path to save
    :param data: text data
    :param encoding: file encoding
    :return: None
    """
    with open(filename, 'w', encoding=encoding) as f:
        f.write(data)


def write_lines(filename, lines, encoding=_ENCODING_UTF8, skip_empty=False, strip=False):
    """
    write lines to file, will add line break for every line automatically
    :param filename: file path to save
    :param lines: lines to save
    :param encoding: file encoding
    :param skip_empty:
    :param strip:
    :return: None
    """
    if isinstance(lines, str):
        raise TypeError('line doesn\'t allow str format')

    if not isinstance(lines, Iterable):
        raise Exception('data can\'t be iterated')

    if strip:
        if skip_empty:
            lines = [l.strip() for l in lines if l.strip()]
        else:
            lines = [l.strip() for l in lines]
    else:
        if skip_empty:
            lines = [l for l in lines if l]

    if not lines:
        raise Exception('lines are empty')

    with open(filename, 'w', encoding=encoding) as f:
        f.write('\n'.join(lines) + '\n')


def read_json(filename):
    """
    read json file
    :param filename: source file path
    :return: loaded object
    """
    with open(filename, encoding=_ENCODING_UTF8) as f:
        return json.load(f)


def write_json(filename, data, serialize_method=None):
    """
    dump json data to file, support non-UTF8 string (will not occur UTF8 hexadecimal code).
    :param filename: destination file path
    :param data: data to be saved
    :param serialize_method: python method to do serialize method
    :return: None
    """
    with open(filename, 'w', encoding=_ENCODING_UTF8) as f:
        if not serialize_method:
            json.dump(data, f, ensure_ascii=False)
        else:
            json.dump(data, f, ensure_ascii=False, default=serialize_method)


def read_jsonline(filename, encoding=_ENCODING_UTF8, default=None):
    """
    read jsonl file
    :param filename: source file path
    :param encoding: file encoding
    :param default: returned value when filename is not existed.
                    If it's None, exception will be raised as usual.
    :return: object list, an object corresponding a line
    """
    if not os.path.exists(filename) and default is not None:
        return default
    file = open(filename, encoding=encoding)
    items = []
    for line in file:
        items.append(json.loads(line))
    file.close()
    return items


def read_jsonline_lazy(filename, encoding=_ENCODING_UTF8, default=None):
    """
    use generator to load jsonl one line every time
    :param filename: source file path
    :param encoding: file encoding
    :param default: returned value when filename is not existed.
                    If it's None, exception will be raised as usual.
    :return: json object
    """
    if not os.path.exists(filename) and default is not None:
        return default
    file = open(filename, encoding=encoding)
    for line in file:
        yield json.loads(line)
    file.close()


def write_jsonline(filename, items, encoding=_ENCODING_UTF8, serialize_method=None):
    """
    write items to file with json line format
    :param filename: destination file path
    :param items: items to be saved line by line
    :param encoding: file encoding
    :param serialize_method: serialization method to process object
    :return: None
    """
    if isinstance(items, str):
        raise TypeError('json object list can\'t be str')

    if not filename.endswith('.jsonl'):
        print('json line filename doesn\'t end with .jsonl')

    if not isinstance(items, Iterable):
        raise TypeError('items can\'t be iterable')
    file = open(filename, 'w', encoding=encoding)
    for item in items:
        file.write(json.dumps(item, ensure_ascii=False, default=serialize_method) + '\n')
    file.close()


def read_ini(filename):
    """
    read configs in ini file
    :param filename: source file path
    :return: parsed config data
    """
    config = configparser.ConfigParser()
    config.read(filename)
    return config


def write_ini(filename, config_data):
    """
    write config into file
    :param filename: destination file
    :param config_data: config data
    :return: None
    """
    config = configparser.ConfigParser()
    for key, val in config_data.items():
        config[key] = val
    with open(filename, 'w') as config_file:
        config.write(config_file)


def append_line(filename, line, encoding=_ENCODING_UTF8):
    """
    append single line to file
    :param filename: destination file path
    :param line: line string
    :param encoding: text encoding to save data
    :return: None
    """
    if not isinstance(line, str):
        raise TypeError('line is not in str type')
    with open(filename, 'a', encoding=encoding) as f:
        f.write(line + '\n')


def append_lines(filename, lines, remove_file=False, encoding=_ENCODING_UTF8):
    """
    append lines to file
    :param filename: destination file path
    :param lines: lines to be saved
    :param remove_file: whether remove the destination file before append
    :param encoding: text encoding to save data
    :return:
    """
    if remove_file and os.path.exists(filename):
        os.remove(filename)
    for line in lines:
        append_line(filename, line, encoding)


def append_jsonline(filename, item, encoding=_ENCODING_UTF8, serialize_method=None):
    """
    append item as a line of json string to file
    :param filename: destination file
    :param item: item to be saved
    :param encoding: file encoding
    :param serialize_method: serialization method to process object
    :return: None
    """
    with open(filename, 'a', encoding=encoding) as f:
        f.write(json.dumps(item, ensure_ascii=False, default=serialize_method) + '\n')


def append_jsonlines(filename, items, encoding=_ENCODING_UTF8, serialize_method=None):
    """
    append item as some lines of json string to file
    :param filename: destination file
    :param items: items to be saved
    :param encoding: file encoding
    :param serialize_method: serialization method to process object
    :return: None
    """
    with open(filename, 'a', encoding=encoding) as f:
        for item in items:
            f.write(json.dumps(item, ensure_ascii=False, default=serialize_method) + '\n')


class __BaseFile(object):
    """
    basic file abstract class, define read, write and append operations
    """

    def __init__(self, filename, encoding, is_remove=False):
        self.filename = filename
        self.encoding = encoding
        if not isinstance(is_remove, bool):
            raise TypeError('is_remove must be bool value')
        if is_remove and os.path.exists(filename):
            os.remove(filename)
        self._file = None
        self.logger = get_logger('FileWrapper')

    def read(self):
        self._to_read()

    def write(self, data):
        self._to_write()

    def append(self, data):
        self._to_append()

    def close(self):
        if self._file is not None and not self._file.closed:
            self._file.close()

    def __change_mode(self, mode):
        """
        change file mode of current file property in the object
        :param mode: new file mode
        :return:
        """
        if not self._file:
            self._file = open(self.filename, mode, encoding=self.encoding)
        elif mode == 'r' or self._file.mode != mode:
            self._file.close()
            self._file = open(self.filename, mode, encoding=self.encoding)

    def _to_read(self):
        if not os.path.exists(self.filename):
            raise FileNotFoundError(self.filename)
        self.__change_mode('r')

    def _to_write(self):
        self.__change_mode('w')

    def _to_append(self):
        self.__change_mode('a')

    def __del__(self):
        self.close()


class TextFile(__BaseFile):
    """
    define raw text operation
    """

    def __init__(self, filename, encoding=_ENCODING_UTF8, is_remove=False):
        super().__init__(filename, encoding, is_remove)

    def read(self):
        self._to_read()
        return self._file.read()

    def read_lines(self, keep_end=False, strip=False, skip_empty=False):
        self._to_read()
        lines = []
        for line in self._file:
            if not keep_end:
                if line[-1] in _LINE_BREAKS:
                    line = line[:-1]
            if strip:
                line = line.strip()
            if skip_empty and not line:
                continue
            lines.append(line)
        return lines

    def write(self, data):
        self._to_write()
        self._file.write(data)

    def write_lines(self, lines):
        self._to_write()
        new_lines = []
        for line in lines:
            if not line.endswith(_LINE_BREAK_TUPLE):
                line += '\n'
            new_lines.append(line)
        self._file.writelines(new_lines)

    def append(self, data):
        self._to_append()
        self._file.write(data)

    def append_line(self, line):
        self._to_append()
        if not line.endswith('\n'):
            line += '\n'
        self._file.write(line)

    def append_lines(self, lines):
        self._to_append()
        for line in lines:
            self.append_line(line)


class JsonLineFile(TextFile):
    """
    define basic operation of jsonline file
    """

    def __init__(self, filename, encoding=_ENCODING_UTF8, is_remove=False):
        super().__init__(filename, encoding, is_remove)

    def read(self):
        return super().read()

    def read_line(self, default=None):
        if not os.path.exists(self.filename) and default is not None:
            return default
        self._to_read()
        for line in self._file:
            yield json.loads(line)

    def read_lines(self, skip_empty=False, default=None, *args, **kwargs):
        if not os.path.exists(self.filename) and default is not None:
            return default
        self._to_read()
        items = []
        for line in self._file:
            items.append(json.loads(line))
        return items

    def write(self, data):
        self._to_write()
        super().write(data)

    def write_line(self, item):
        self._to_write()
        self._file.write(self._to_string(item))

    def write_lines(self, lines):
        for line in lines:
            self.write_line(line)

    def append(self, data):
        self._to_append()
        self._file.write(self._to_string(data))

    def append_line(self, line):
        self.append(line)

    def append_lines(self, lines):
        for line in lines:
            self.append_line(line)

    def _to_string(self, item, append_line_break=True):
        if not isinstance(item, str):
            item = json.dumps(item, ensure_ascii=False)
        if append_line_break and not item.endswith('\n'):
            item += '\n'
        return item
